
    #   Render sky with planets in conjunction from
    #	an extract file.  This is usually run from an extract
    #	with a single integration step, such as with the demo:
    #
    #	    perl conjsky.pl 86859_bary_extract.csv
    #
    #	It leaves the .pov file and the rendering in the
    #	current directory so you can twiddle with them.
    #	Note that if Jupiter isn't involved in the conjunction,
    #	you'll need to change $target_planet below to whichever
    #	planet you wish centred in the rendering.
    
    #	by John Walker
    #	http://www.fourmilab.ch/
    
    #	For details, please see: http://www.fourmilab.ch/documents/canon_transits/

    use strict;
    use warnings;
    
    use Math::VectorReal;
    use Math::Trig;
    
    #	Observer's planet.  Specify "Sun" for heliocentric
    my $observer_planet = 'Sun';
    
    #	Target planet (will be centred in rendering)
    my $target_planet = 'Jupiter';
    
    #	Horizontal field of view angle in degrees.
    #	Vertical view angle will be 3/4 this value.
    my $vangle = 15;
    
    #	Inflate the sizes of planets by this factor to render them more visible
    my $inflate_planets = 200;
    
    #	Specify animation frame size
    my ($fwidth, $fheight) = (1024, 768);
    
    my $pvopts = "-v -w$fwidth -h$fheight -d -q9";	# POV-Ray rendering options
    
    my $au = 149597870691.0;	    	# Astronomical unit in metres
        
    my @planets = (
    	"Sun",	    	    #  0
    	"Mercury",  	    #  1
	"Venus",    	    #  2
    	"Earth",    	    #  3
	"Moon",     	    #  4
	"Mars",     	    #  5
	"Jupiter",  	    #  6
	"Saturn",   	    #  7
	"Uranus",   	    #  8
     	"Neptune",  	    #  9
	"Pluto" );  	    # 10

    my %planet_col = (
    	"Sun", "White",
    	"Mercury", "Brown",
	"Venus", "Red",
    	"Earth", "Orange",
	"Moon", "Gray75",
	"Mars", "Yellow",
	"Jupiter", "Orange",
	"Saturn", "Yellow",
	"Uranus", "Cyan",
     	"Neptune", "Blue",
	"Pluto", "Yellow" );
	
    my %planet_radius = (
    	"Sun", 6.96e8 / $au,
    	"Mercury", 2439700 / $au,
	"Venus", 6051800 / $au,
    	"Earth", 4.26352325064817808471E-5,
	"Moon", 1.16178124180819699985E-5,
	"Mars", 3397000 / $au,
	"Jupiter", 71492000 / $au,
	"Saturn", 60268000 / $au,
	"Uranus", 25559000 / $au,
     	"Neptune", 24764000 / $au,
	"Pluto", 1195000 / $au );
	
    my $l;
    
    my ($jd, $yy, $mo, $dd, $hh, $mm, $ss);
    my %planet_vec;
    my $ijd = 0;
    
    while ($l = <>) {
	#   Load a state vector snapshot into memory
	
	$l =~ m/JD,([^,]+),([^,]+),([^,]+),([^,]+),([^,]+),([^,]+),([^,]+)\s/;
    	($jd, $yy, $mo, $dd, $hh, $mm, $ss) = ($1, $2, $3, $4, $5, $6, $7);
	
	while (1) {
	    $l = <>;
	    if ($l =~ m/^END/) {
	    	last;
	    }
	    $l =~ m/^(\w+),([\d\-\.e]+),([\d\-\.e]+),([\d\-\.e]+)/;
	    $planet_vec{$1} = vector($2, $3, $4);
	}
	
	objangles();
	
    	my $fname = sprintf("conjsky_%.2f.pov", $jd);
	genscene($fname);
	renderscene($fname);
    }
    
    #	Compute angles between objects in conjunction
    
    sub objangles {
    	print("Angular Separation of Planet Pairs\n\n");
	
    	print("Planet 1   Planet 2    Angle\n");
    	print("---------  ---------  -------\n");
    	for (my $i = 0; $i <= $#planets - 1; $i++) {
	    if ($planets[$i] ne $observer_planet) {
	    	for (my $j = $i + 1; $j <= $#planets; $j++) {
		    if ($planets[$j] ne $observer_planet) {
		    	my $p1vec = ($planet_vec{$planets[$i]} - $planet_vec{$observer_planet})->norm;
		    	my $p2vec = ($planet_vec{$planets[$j]} - $planet_vec{$observer_planet})->norm;
			my $ang = rad2deg(asin(($p1vec x $p2vec)->length));
			if (($p1vec . $p2vec) < 0) {
			    $ang += 180;
			}
			printf("%-8s   %-8s   %6.2f\n", $planets[$i], $planets[$j], $ang);
		    }
		}
	    }
	}
	print("\n");
    }
    	
    sub genscene {
    	my ($fname, $vfi) = @_;
	
	my $observer_pos = outpos($planet_vec{$observer_planet});
	my $target_pos = outpos($planet_vec{$target_planet});
	my $sun_pos =  outpos($planet_vec{'Sun'});
	my $view_direction = ($planet_vec{$target_planet} - $planet_vec{$observer_planet})->norm;
	
	open(OF, ">$fname");
	
	print OF << "EOD";

#include "colors.inc"

    camera {
    	angle $vangle
	location $observer_pos  // $observer_planet
	sky <0, 1, 0>
	look_at $target_pos  // $target_planet
    }

    light_source {
    	$observer_pos  // Sun
	color White
    }
    
EOD

    for (my $i = 0; $i <= 10; $i++) {
    	my $pname = $planets[$i];
	if (($pname ne 'Moon') && ($pname ne $observer_planet)) {
	    my $ppos = outpos($planet_vec{$pname});
	    my $psize = $planet_radius{$pname} * (($i > 0) ? $inflate_planets : 1);
	    my $orbital_radius = $planet_vec{$pname}->length;
	    
	    print OF << "EOD";
    	//  $pname
    	sphere {
      	    $ppos, $psize
	    texture {
		pigment {
	    	    color $planet_col{$pname}
		}
	    }
    	}
    
EOD
 	}
    }
    
    close(OF);
	
    }
    
    #	Render the scene for a given instant
    
    sub renderscene {
    	my ($fname) = @_;
	
	my $ofn = $fname;
	$ofn =~ s/pov$/png/;
	my $flabel = sprintf("%.2f  %4d-%02d-%02d %02d:%02d", $jd, $yy, $mo, $dd, $hh, $mm);
	my $ly = $fheight - 16;
	my $c = "povray $fname $pvopts +fp +o- | " .
		"ppmlabel -x 10 -y $ly -size 10 -text '$flabel' | " .
		"pnmtopng >$ofn\n";
	command($c);
    }
    
    #	Edit a vector to POV-Ray format

    sub outpos {
    	my ($p) = @_;
	
	return "< " . $p->x() . ", " . $p->y() . ", " . $p->z(). " >";
    }

    #	Print a command and then run it
    
    sub command {
    	my ($c) = @_;
	
	print($c);
	system($c);
    }
